#****
# [manpage_begin message n 0.5]
# [moddesc {Message handling for Mailroom}]
# [copyright {Copyright 2009 Jeremy Cowgar. All Rights Reserved.}]
# [category {Message Handling}]
# [require mime]
# [require md4]
# [keywords mailroom message mime]
# [description]
# Provides procedures for maintaining the message database in Mailroom.
#

namespace eval ::mailroom::message {
  package require mime
  package require md4
  
  #****
  # [section Commands]
  #
  # [list_begin definitions]
  #

  #****
  #
  # [call [cmd {::mailroom::message send}] \
  # [opt "[option -string] [arg string]"] \
  # [opt "[option -mime] [arg mime-message]"] \
  # [opt "[option -headers] [arg list]"] \
  # [opt "[option -to] [arg address-list]"] \
  # [opt "[option -cc] [arg address-list]"] \
  # [opt "[option -bcc] [arg address-list]"] \
  # [opt "[option -insert] [arg boolean]"] \
  # [opt "[option -statusCmd] [arg procname]"] \
  # ]
  #
  # Send a message via the configured SMTP server.
  #
  # [list_begin definitions]
  # [def "[option -string] [arg string]"]
  # The message content as an unparsed string.
  #
  # [def "[option -mime] [arg mime-message]"]
  # The message content as a parsed mime message. Mime messages passed will not automatically be finalized,
  # this is the responsibility of the caller.
  #
  # [def "[option -headers] [arg header-list]"]
  # A list of additional headers that will be added to the message before sending. The original mime message,
  # if supplied, will not be altered.
  #
  # [def "[option -to] [arg address-list]"]
  # An email list of who to send the email to.
  #
  # [def "[option -cc] [arg address-list]"]
  # An email list of who to carbon-copy the email to.
  #
  # [def "[option -bcc] [arg address-list]"]
  # An email list of who to blind carbon-copy the email to.
  #
  # [def "[option -insert] [arg boolean]"]
  # Defauls to 1. If true, the message will be inserted into the mailroom message database.
  #
  # [def "[option -bucket] [arg name]"]
  # Defaults to inbox.
  #
  # [def "[option -statusCmd] [arg procname]"]
  # If supplied, it is called at key points through the send process with a string message as to the current
  # send state.
  #
  # [list_end]
  #
  proc send {args} {
    set insert 1
    set msg {}
    set fromString 0
    set headers {}
    set toAddrs {}
    set ccAddrs {}
    set bccAddrs {}
    set bucket inbox
    set statusCmd 0
    
    foreach {arg val} $args {
      switch -- $arg {
        -insert    { set insert $val }
        -headers   { set headers $val }
        -to        { set toAddrs $val }
        -cc        { set ccAddrs $val }
        -bcc       { set bccAddrs $val }
        -bucket    { set bucket $val }
        -statusCmd { set statusCmd $val }
        -string {
          set msg [mime::initialize -canonical text/plain -string $val]
          set fromString 1
        }
        -mime {
          set msg $val
          if {$headers == {}} {
            set headers [mime::getheader $msg]
          }
        }
        default {
          error "Invalid argument `$arg` passed to send. Must be one of -insert"
        }
      }
    }
    
    # Ensure we were given a message
    if {$msg=={}} {
      error "no message passed to insert. see `-string` or `-mime`."
    }
    
    # Add our To, Cc and From headers.
    lappend headers From [::mailroom::full_from]
    if {$toAddrs != {}} {
      lappend headers To [join $toAddrs {, }]
    }
    if {$ccAddrs != {}} {
      lappend headers Cc [join $ccAddrs {, }]
    }
    
    # Add misc headers
    lappend headers X-Mailer "$::mailroom::appTitle v$::mailroom::appVersion ($::mailroom::appUrl)"
    lappend headers User-Agent "$::mailroom::appTitle v$::mailroom::appVersion ($::mailroom::appUrl)"
    if {[lsearch -nocase $headers date] == -1} {
      lappend headers Date [clock format [clock seconds] -format {%a, %d %b %Y %H:%M:%S %z}]
    }
    
    # Ensure we have a Message-ID
    if {[lsearch -nocase $headers message-id]==-1} {
      lappend headers Message-ID [new_uuid]
    }
    
    array set cfg [::mailroom::db eval {
        SELECT name2, value FROM var
        WHERE name1='config' AND name2 LIKE 'smtp%'
      }]
    
    set optParams {}
    if {$cfg(smtptls)} {
      lappend optParams -usetls
      lappend optParams 1
    }
    if {$cfg(smtpuserid)!=""} {
      lappend optParams -username
      lappend optParams $cfg(smtpuserid)
    }
    if {$cfg(smtppasswd)!=""} {
      lappend optParams -password
      lappend optParams $cfg(smtppasswd)
    }
    
    set smtpHeaders {}
    foreach {h v} $headers {
      lappend smtpHeaders -header [list $h $v]
    }
    
    if {$statusCmd != 0} {
      $statusCmd "Sending message"
    }

    set cmd [list smtp::sendmessage $msg -servers [list $cfg(smtphost)] -ports [list $cfg(smtpport)] \
      {*}$smtpHeaders {*}$optParams]
    
    log::log debug $cmd
    
    log::log notice "Sending message via $cfg(smtphost):$cfg(smtpport)"
    
    # TODO: Capture error, set the message in an unsent state if an error occurs
    eval $cmd
    
    log::log notice "Sent"
    
    if {$insert} {
      # Add the headers to the mime object for standardization on insert proc
      foreach {h v} $headers {
        mime::setheader $msg $h $v
      }
      
      if {$statusCmd != 0} {
        $statusCmd "Saving message"
      }
      
      log::log notice "Saving message into bucket $bucket"
      
      insert -mime $msg -is_read 1 -bucket $bucket
    }
    
    if {$fromString} {
      mime::finalize $msg
    }

    if {$statusCmd != 0} {
      $statusCmd "Message sent"
    }

    return 1
  }

  #****
  # [call [cmd {::mailroom::message get}] [arg id] [arg array_name]]
  #
  # Get a message, identified by [arg id], from the database and place it's contents into the array named
  # by [arg array_name].
  #
  proc get {id array_name} {
    upvar 1 $array_name x
    
    ::mailroom::db eval {
      SELECT
        m.bucket,
        datetime(m.received) AS received,
        datetime(m.sent) AS sent,
        m.size,
        m.attachment_count,
        m.subject,
        m.is_read,
        b.data AS b_data,
        f.userid AS from_userid,
        f.address AS from_address,
        f.fulladdr AS from_full_address,
        datetime(f.recent) AS from_recent,
        f.to_count AS from_to_count,
        f.from_count AS from_from_count,
        f.spam_count AS from_spam_count,
        f.edited AS from_edited
      FROM
        msg AS m
        JOIN blob AS b ON (b.blobid=m.msgid)
        JOIN user AS f ON (m.from_userid=f.userid)
      WHERE
        m.msgid=$id
    } x break
  }

  #****
  #
  # [call [cmd {::mailroom::message insert}] \
  # [opt "[option -string] [arg string]"] \
  # [opt "[option -mime] [arg mime-message]"] \
  # [opt "[option -is_read] [arg boolean]"] \
  # [opt "[option -receivedTime] [arg integer]"] \
  # [opt "[option -bucket] [arg name]"] \
  # ]
  #
  # Insert a message into the database. You must specify the message content with either the [option -string]
  # or [option -mime] arguments.
  #
  # [list_begin definitions]
  # [def "[option -string] [arg string]"]
  # Message content as an unparsed string.
  #
  # [def "[option -mime] [arg mime-message]"]
  # Message content as an already parsed mime message.
  #
  # [def "[option -is_read] [arg boolean]"]
  # Defaults to 0. Boolean value indicating if the message has been read or not.
  #
  # [def "[option -receivedTime] [arg integer]"]
  # Defaults to \[clock seconds]. Clock value of when the message was received if other than right now. If
  # [option -receivedTime] is `0`, then the receive time is taken from the email message and will be the
  # same as the Date: header field.
  #
  # [def "[option -bucket] [arg name]"]
  # Defaults to "inbox". Bucket to insert the message into.
  # [list_end]
  #
  proc insert {args} {
    set receivedTime {}
    set is_read 0
    set bucket inbox
    set msg {}
    set m {}
    set fromString 0
    
    foreach {arg val} $args {
      switch -- $arg {
        -is_read { set is_read $val }
        -receivedTime { set receivedTime $val }
        -bucket { set bucket $val }
        -string {
          set msg $val
          set m [_mime_from_string $msg]
          set fromString 1
        }
        -mime {
          set m $val
          set msg [mime::buildmessage $m]
        }
        default {
          error "Invalid argument `$arg` given to insert"
        }
      }
    }
    
    if {$m == {}} {
      error "Message was not passed to insert. See `-string` or `-mime`."
    }
  
    set hdrTo [_header $m to]
    set hdrCc [_header $m cc]
    set hdrFrom [_header $m from 0]
    set hdrDate [_header $m date 0]
    set hdrReferences [_header $m references]
    set hdrSubject [_header $m subject 0]
    set hdrReceived [_header $m received]
    set hdrMessageId [_header $m message-id 0]
    
    array set px [mime::getproperty $m]
    
    set to [join [concat $hdrTo $hdrCc]]
    set from $hdrFrom
    array set fromAddr [lindex [mime::parseaddress $hdrFrom] 0]
    set subject [_subject_fix $hdrSubject]
    if {$hdrMessageId == {}} {
      set hdrMessageId [md4::md4 -hex [mime::getheader $m]]
    }
    
    set oldMsgId [::mailroom::db one {SELECT msgid FROM msg_uuid WHERE uuid=$hdrMessageId}]
    if {$oldMsgId != ""} {
      ::mailroom::db transaction {
        ::mailroom::db eval {UPDATE blob SET data=$msg WHERE blobid=$oldMsgId}
        ::mailroom::db eval {UPDATE msg SET subject=$subject WHERE msgid=$oldMsgId}
      }
        
      ::log::log notice "Upd: $subject"
      
      if {$fromString} {
        mime::finalize $m
      }
      
      return 0
    }
    
    set body [_body $m]
    
    if {[catch {clock scan $hdrDate} sent]} {
      set sent [::mailroom::julianday]
    } else {
      set sent [::mailroom::julianday $sent]
    }
    
    if {$receivedTime == {}} {
      set receivedTime [::mailroom::julianday]
    } elseif {$receivedTime == 0} {
      set receivedTime $sent
    }
    
    set toAddrs [join $hdrTo]
    set ccAddrs [join $hdrCc]
    
    if {![info exists px(parts)]} {
      set attachmentCount 0
    } else {
      set attachmentCount [expr {[llength $px(parts)] - 1}]
    }
    set messageSize [string length $msg]
    
    ::mailroom::db transaction {
      ::mailroom::db eval {
        INSERT INTO blob (data) VALUES ($msg)
      }
      set blobId [::mailroom::db last_insert_rowid]
      
      set fromUserId [::mailroom::user new_from $from $receivedTime]
      
      set from [string map {< {&lt;} > {&gt;} & {&amp;}} $from]
      set to [string map {< {&lt;} > {&gt;} & {&amp;}} $to]
      set subject [string map {< {&lt;} > {&gt;} & {&amp;}} $subject]
      set body [string map {< {&lt;} > {&gt;} & {&amp;}} $body]
      ::mailroom::db eval {
        INSERT INTO email (rowid, from_who, to_who, subject, body)
        VALUES ($blobId, $from, $to, $subject, $body)
      }
      
      # If the from address is the same as our from address, we sent it.
      if {$fromAddr(address) == [::mailroom::config option config fromaddr]} {
        foreach addr [concat $hdrTo $hdrCc] {
          ::mailroom::user new_to $addr -time $receivedTime
        }
        
        set is_read 1
      }
      
      ::mailroom::db eval {
        INSERT INTO msg (
          msgid, bucket, received, sent, size, attachment_count, from_userid, subject, is_read
        ) VALUES (
          $blobId, $bucket, $receivedTime, $sent, $messageSize, $attachmentCount,
          $fromUserId, $subject, $is_read
        );
        INSERT INTO msg_uuid (msgid, uuid) VALUES ($blobId, $hdrMessageId);
      }
      
      ::mailroom::user callback after_insert $blobId $m
    }
    
    if {$fromString} {
      mime::finalize $m
    }
    
    log::log notice "Ins: $subject"
    
    return 1
  }

  #****
  # [call [cmd {::mailroom::message delete}] [arg id]]
  #
  # Delete the message identified by [arg id].
  #
  proc delete {id} {
    ::mailroom::db transaction {
      ::mailroom::db eval {SELECT blobid FROM msg WHERE msgid=$msgid} break
      if {![info exists blobid]} return
      ::mailroom::db eval {
        DELETE FROM blob WHERE blobid=$blobid;
        DELETE FROM email WHERE rowid=$msgid;
        DELETE FROM msg_uuid WHERE msgid=$msgid;
        DELETE FROM msg_send_queue WHERE msgid=$msgid;
        DELETE FROM msg WHERE msgid=$msgid;
      }
    }
  }
  
  #****
  # [call [cmd {::mailroom::message delete_by_bucket}] [arg name]]
  #
  # Delete all messages in the [arg name] bucket. [arg name] can also be a list of buckets to delete messages
  # in.
  #
  proc delete_by_bucket {name} {
    set dellist ('[join $name ',']')
    
    ::mailroom::db transaction {
      ::mailroom::db eval "CREATE TEMP TABLE todel AS
        SELECT msgid FROM msg WHERE bucket IN $dellist"
      ::mailroom::db eval {
        DELETE FROM blob WHERE blobid IN todel;
        DELETE FROM email WHERE rowid IN todel;
        DELETE FROM msg_uuid WHERE msgid IN todel;
        DELETE FROM msg_send_queue WHERE msgid IN todel;
        DELETE FROM msg WHERE msgid IN todel;
        DROP TABLE todel;
      }
    }
  }
  
  #****
  # [call [cmd {::mailroom::message set_bucket}] \
  # [arg messageId] \
  # [arg newBucket] \
  # [opt "[option -markRead] [arg boolean]"] \
  # ]
  #
  # Set the bucket of message [arg message-id] to [arg newBucket]. [option -markRead] defaults to 0.
  #
  proc set_bucket {messageId newBucket args} {
    set markRead 0
    
    foreach {a v} $args {
      switch -- $a {
        -markRead { set markRead $v }
        default {
          error "Invalid option `$a` sent to set_bucket. Must be one of `-markRead`"
        }
      }
    }
    
    ::mailroom::db eval {SELECT from_userid, bucket AS old_bucket FROM msg WHERE msgid=$messageId} {
      if {$old_bucket != "spam" && $newBucket == "spam"} {
        ::mailroom::user inc $from_userid spam_count
      } elseif {$old_bucket == "spam" && $newBucket != "spam"} {
        ::mailroom::user inc $from_userid spam_count -1
      }
      ::mailroom::db eval {UPDATE msg SET bucket=$newBucket WHERE msgid=$messageId}
      if {$markRead} {
        ::mailroom::db eval {UPDATE msg SET is_read=1 WHERE msgid=$messageId}
      }
    }
  }
  
  #****
  # [call [cmd {::mailroom::message new_uuid}]]
  #
  # Create a unique message uuid
  #
  proc new_uuid {} {
    set sender [::mailroom::config option config fromaddr]
    if {$sender == ""} {
      set sender [info hostname]
    } else {
      regsub {^.*@} $sender {} sender
    }
    
    set uuid [::mailroom::db onecolumn {SELECT julianday('now') || abs(random()) || abs(random())}]
    append uuid @$sender
    
    return $uuid
  }
  
  #****
  # [call [cmd {::mailroom::message import_file}] [arg filename] \
  # [opt "[option -statusCmd] [arg procname]"] \
  # ]
  #
  # Import a file into our database. [arg filename] is considered to be in the MBOX file format as described
  # by QMail (http://www.qmail.org/man/man5/mbox.html).
  #
  proc import_file {filename args} {
    set statusCmd 0
    
    foreach {a v} $args {
      switch -- $a {
        -statusCmd { set statusCmd $v }
        default {
          error "Invalid option sent to import_file. Must be one of `-statusCmd`"
        }
      }
    }
    
    set in [open $filename]
    fconfigure $in -translation binary
    set buffer {}
    
    set importCount 0
    set lineNo 0
    if {$statusCmd != 0} {
      $statusCmd "Importing..."
    }
    ::mailroom::db transaction {
      while {![eof $in]} {
        set line [gets $in]
        incr lineNo
        if {[regexp {^From } $line]} {
          if {[string length $buffer] > 0} {
            log::log debug "Importing message #$importCount"
            if {[catch {::mailroom::message insert -string $buffer -bucket inbox -receivedTime 0} msg]} {
              log::log error "Could not import message $importCount, line number: $lineNo"
              log::log error "Error message: $msg"
              log::log error "Message was skipped!"
            } else {
              incr importCount
              if {$statusCmd != 0} {
                $statusCmd "Imported #$importCount"
              }
            }
          }
          set buffer "X-From: [string range $line 5 end]\n"
        } else {
          if {[regexp {^>From } $line]} {
            set line [string range $line 1 end]
          }
          append buffer "$line\n"
        }
      }
      close $in
      
      if {[string length $buffer] > 0} {
        log::log debug "Importing message #$importCount"
        insert -string $buffer -bucket inbox -receivedTime 0
        incr importCount
      }
    }
    
    if {$statusCmd != 0} {
      $statusCmd "$importCount msgs imported."
    }
  }
  
  #
  # Import a database into our database
  #
  # Arguments:
  #    -db DATABASE_FILENAME
  #
  proc import_db {args} {
    set filename {}
    
    foreach {arg val} $args {
      switch -- $arg {
        -db { set filename $val }
        default {
          error "Invalid argument `$arg` given to import_file. Must be one of -file"
        }
      }
    }
    
    error "import_db is not yet implemented"
  }
    
  #####
  #
  # Supporting Routines
  #
  ####
  
  proc _mime_from_string {s} {
    if {[catch {mime::initialize -string $s} m]} {
      regsub -nocase {Content-Type: multipart/[^; ]*} $s \
        {Content-Type: text/plain} s
      set m [mime::initialize -string $s]
    }
    
    return $m
  }
  
  # Clean up a subject line.
  #
  # Do RFC 2047 decoding.
  #
  proc _subject_fix {orig {strip_prefixes 0}} {
    regsub -all {\s+} [mime::field_decode $orig] { } x1
    if {$strip_prefixes==0} {
      return $x1
    }
  
    regsub -all {R[eE](\[\d+\])?: +} $x1 {} x2
    regsub -all {F[wW][dD]?(\[\d+\])?: +} $x2 {} x3
  
    return $x3
  }
  
  #
  # Get a header from a mime message or return an empty string
  #
  proc _header {msg hdr {idx -1}} {
    if {[catch {mime::getheader $msg $hdr} result]} {
      set result {}
    }
    
    if {$idx > -1} {
      return [lindex $result $idx]
    }
    
    return $result
  }
  
  # Extract and return the body of a message from a mime token.  If html is 0,
  # the body is the first part we come to that has a content type of text/plain
  # is returned as the body, otherwise the first text/html content is returned
  # as the body. This routine recursively decends the mime hierarchy of the
  # message until it finds such a part.  If no valid body is found an empty string
  # is returned.
  #
  proc _body {m {html 0}} {
    array set px [mime::getproperty $m]
    set body {}
    if {![info exists px(parts)]} {
      set body [mime::getbody $m]
      if {[info exists px(content)]} {
        if {$px(content)=="text/html"} {
          if {!$html} {
            regsub -nocase -all {<(style[^>]*>)?[^>]+>} $body {} body
            regsub -nocase -all {\s*\n\s*\n\s*} $body "\n\n" body
          }
        } elseif {$html} {
          set body {}
        } elseif {$px(content)!="text/plain"} {
          set body [string map {< {&lt;} > {&gt;} & {&amp;}} $body]
        }
      }
    } else {
      foreach part $px(parts) {
        set body [_body $part $html]
        if {$body!=""} break
      }
    }
    return $body
  }
  
  #****
  # [list_end]
  #
  # [section Examples]
  #
  # [section Todo]
  #
  # [section {Bugs, Ideas, Feedback}]
  # Provide feedback to Jeremy Cowgar <jeremy@cowgar.com>
  #
  
  namespace export send get insert delete delete_by_bucket set_bucket new_uuid import_file
  namespace ensemble create
}

#****
# [manpage_end]
